﻿using System;
using System.Text;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using BeanboxSoftware.BeanMap;

namespace BeanMapTest
{
  [TestClass]
  public class IndexNotationTest
  {
    class KeyTypeOne
    {
    }

    class KeyTypeTwo
    {
    }

    class ValueType
    {
    }

    [TestMethod]
    public void JaggedAccessTest()
    {
      var map = new Map<KeyTypeOne, KeyTypeTwo, ValueType>();

      var a = new KeyTypeOne();
      var i = new KeyTypeTwo();
      var v = new ValueType();

      var child = map[a];

      Assert.IsFalse(map.ContainsKey(a));
      Assert.IsNotNull(child);

      child[i] = v;

      Assert.IsTrue(map.ContainsKey(a));

      Assert.AreEqual(v, map[a][i]);
      Assert.AreEqual(v, map[a, i]);

      map.Clear();
      Assert.IsFalse(map.ContainsKey(a));
      Assert.IsFalse(map.ContainsKey(a, i));
      map[a][i] = v;
      Assert.AreEqual(v, map[a][i]);
      Assert.AreEqual(v, map[a, i]);
      Assert.AreEqual(v, child[i]);
    }

    [TestMethod]
    public void JaggedConfigTest()
    {
      var map = new Map<KeyTypeOne, KeyTypeTwo, ValueType>();

      var a = new KeyTypeOne();
      var i = new KeyTypeTwo();
      var v = new ValueType();
      var e = new ValueType();

      map.DefaultValue = e;

      var child = map[a];

      Assert.AreEqual(e, map[a][i]);
      Assert.AreEqual(e, map[a, i]);
      Assert.AreEqual(e, child[i]);
      Assert.IsFalse(map.ContainsKey(a, i));
      Assert.IsFalse(child.ContainsKey(i));

      map.DefaultGeneration = (k1, k2) => v;

      Assert.AreEqual(v, map[a][i]);

      Assert.IsTrue(map.ContainsKey(a, i));
      Assert.IsTrue(child.ContainsKey(i));

      map.Clear();
      Assert.IsFalse(map.ContainsKey(a, i));
      Assert.AreEqual(v, map[a, i]);
      
      Assert.IsTrue(map.ContainsKey(a, i));
      Assert.IsTrue(child.ContainsKey(i));
    }
  }
}
